import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import convert_config_to_qualified_name
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .utils import set_default


class RecordClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(
        metadata={"doc": "Set true to generate the class (default)"}
    )
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "Model base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to copy to record file"},
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()
    fields = ma.fields.Dict(
        attribute="fields",
        data_key="fields",
        metadata={"doc": "Extra fields to add to the class"},
    )


class RecordModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        record = ma.fields.Nested(
            RecordClassSchema, metadata={"doc": "api/Record settings"}
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        module = datatype.definition["module"]["qualified"]
        profile_module = context["profile_module"]
        record_prefix = datatype.definition["module"]["prefix"]

        record = set_default(datatype, "record", {})
        record.setdefault("generate", True)
        records_module = record.setdefault("module", f"{module}.{profile_module}.api")
        record.setdefault("class", f"{records_module}.{record_prefix}Record")
        record.setdefault(
            "base-classes",
            ["invenio_records_resources.records.api.Record{InvenioRecord}"],
        )
        record.setdefault(
            "imports",
            [],
        )
        record.setdefault("extra-code", "")
        record.setdefault("fields", {})
        convert_config_to_qualified_name(record)



import dataclasses
from typing import Dict, List, Optional

from oarepo_model_builder.utils.jinja import (
    extract_extra_code_imports,
    generate_extra_code,
)
from oarepo_model_builder.utils.python_name import PythonQualifiedName


@dataclasses.dataclass
class FacetDefinition:
    path: str
    dot_path: str
    searchable: bool
    imports: List[Dict[str, str]]
    facet_groups: Dict[str, int]
    facet: Optional[bool]
    field: Optional[str] = None

    def update(self, facet_section):
        self.imports.extend(facet_section.get("imports", []))
        if self.searchable is None:
            # searchable not set up, so take it from argument
            self.searchable = facet_section.get("searchable", None)

    def set_field(self, facet_section, arguments, field_class=None):
        field = facet_section.get("field")
        extra_imports = None
        if field:
            extra_imports = extract_extra_code_imports(field)
            self.field = generate_extra_code(field)
        else:
            if not field_class:
                field_class = facet_section.get("facet-class")
            if field_class:
                field_class = PythonQualifiedName(field_class)
                self.field = f"{field_class.local_name}({', '.join(arguments)})"
                extra_imports = field_class.imports

        if extra_imports:
            self.imports = [*self.imports, *extra_imports]



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import convert_config_to_qualified_name
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .utils import set_default


class ResourceClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(metadata={"doc": "Generate the resource class (default)"})
    config_key = ma.fields.Str(
        metadata={
            "doc": "Name of the config entry that holds the current resource class name"
        }
    )
    proxy = ma.fields.Str(
        metadata={"doc": "name of the generated proxy, will be put to _proxies_ module"}
    )
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the resource class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "A list of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be put to the bottom of the generated file"},
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    additional_args = ma.fields.List(
        ma.fields.String(),
        attribute="additional-args",
        data_key="additional-args",
        metadata={
            "doc": "List of additional arguments that will be passed to the resource constructor"
        },
    )
    skip = ma.fields.Boolean()


class ResourceConfigClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    base_url = ma.fields.String(
        data_key="base-url",
        attribute="base-url",
        metadata={"doc": "The base url of the resource"},
    )

    base_html_url = ma.fields.String(
        data_key="base-html-url",
        attribute="base-html-url",
        metadata={"doc": "The base html url of the resource"},
    )

    generate = ma.fields.Bool(
        metadata={"doc": "Generate the resource config class (default)"}
    )
    config_key = ma.fields.Str(
        metadata={
            "doc": "Name of the config entry that holds the current resource class name"
        }
    )
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the config class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "A list of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be put below the generated config class"},
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    additional_args = ma.fields.List(
        ma.fields.String(),
        attribute="additional-args",
        data_key="additional-args",
        metadata={
            "doc": "List of additional arguments that will be passed to the resource config constructor"
        },
    )
    skip = ma.fields.Boolean()


class ResourceModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        resource = ma.fields.Nested(
            ResourceClassSchema, metadata={"doc": "Resource class settings"}
        )
        resource_config = ma.fields.Nested(
            ResourceConfigClassSchema,
            attribute="resource-config",
            data_key="resource-config",
            metadata={"doc": "Resource config class settings"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        module_container = datatype.definition["module"]
        profile_package = context["profile_module"]

        resource_package = (
            f"{module_container['qualified']}.resources.{profile_package}"
        )

        resource = set_default(datatype, "resource", {})
        resource.setdefault("generate", True)
        resource.setdefault(
            "config-key",
            f"{module_container['base-upper']}_{context['profile_upper']}_RESOURCE_CLASS",
        )
        resource_module = resource.setdefault("module", f"{resource_package}.resource")
        resource.setdefault(
            "class",
            f"{resource_module}.{module_container['prefix']}Resource",
        )
        resource.setdefault(
            "proxy",
            "current_resource",
        )
        resource.setdefault("extra-code", "")
        resource.setdefault(
            "base-classes", ["invenio_records_resources.resources.RecordResource"]
        )
        resource.setdefault(
            "imports",
            [],
        )
        resource.setdefault("additional-args", [])
        convert_config_to_qualified_name(resource)

        config = set_default(datatype, "resource-config", {})
        config.setdefault("generate", True)
        config.setdefault("base-url", f"/{module_container['kebab-module']}/")
        config.setdefault("base-html-url", f"/{module_container['kebab-module']}/")
        config.setdefault(
            "config-key",
            f"{module_container['base-upper']}_{context['profile_upper']}_RESOURCE_CONFIG",
        )
        config_module = config.setdefault("module", f"{resource_package}.config")
        config.setdefault(
            "class",
            f"{config_module}.{module_container['prefix']}ResourceConfig",
        )
        config.setdefault("extra-code", "")

        config.setdefault(
            "base-classes", ["invenio_records_resources.resources.RecordResourceConfig"]
        )
        config.setdefault(
            "imports",
            [],
        )
        config.setdefault("additional-args", [])
        convert_config_to_qualified_name(config)



import copy
import os
import pathlib
from pathlib import Path
from typing import Callable, Dict, List, Union

from jsonpointer import resolve_pointer

from .exceptions import IncludedFileNotFoundException
from .utils.deepmerge import deepmerge
from .validation import validate_model


class ModelSchema:
    USE_KEYWORD = "use"
    REF_KEYWORD = "$ref"
    EXTEND_KEYWORD = "extend"

    def __init__(
        self,
        file_path,
        content=None,
        included_models: Dict[str, Callable] = None,
        merged_models: List[Union[str, Path]] = None,
        loaders=None,
        validate=True,
        source_locations=None,
        reference_processors=None,
        post_reference_processors=None,
    ):
        """
        Creates and parses model schema

        :param file_paths: list of paths on the filesystem to the model schema files. The content of these files will be merged before the processing.
        :param content:   if set, use this content, otherwise load the file_path
        :param included_models: a dictionary of file_id to callable that returns included json.
                The callable expects a single parameter, an instance of this schema
        """

        self.file_path = file_path
        self.included_schemas = included_models or {}
        self.loaders = loaders
        self.source_locations = [*(source_locations or [])]
        self.source_locations.append(os.path.abspath(os.curdir))
        self.source_locations.append(os.path.dirname(os.path.abspath(self.file_path)))
        self._reference_processors = deepmerge(
            {
                self.REF_KEYWORD: [],
                self.USE_KEYWORD: [],
                self.EXTEND_KEYWORD: [],
            },
            reference_processors or {},
        )
        self._post_reference_processors = deepmerge(
            {
                self.REF_KEYWORD: [],
                self.USE_KEYWORD: [],
                self.EXTEND_KEYWORD: [],
            },
            post_reference_processors or {},
        )

        if content is not None:
            self.schema = content
        else:
            self.schema = copy.deepcopy(self._load(file_path))
            for fp in merged_models or []:
                self.schema = deepmerge(self.schema, copy.deepcopy(self._load(fp)))

        reference_count = 1
        while reference_count:
            reference_count = self._resolve_references(
                self.schema,
                [],
                {
                    self.REF_KEYWORD,
                    self.USE_KEYWORD,
                },
            )
            reference_count += self._resolve_references(
                self.schema,
                [],
                {
                    self.EXTEND_KEYWORD,
                },
            )

        self._resolve_shortcuts(self.schema)

        # any star keys should be kept
        use_star_keys(self.schema)

        self.schema.setdefault("settings", {})

        self._sections = {}

        self._strip_ignored_elements()

        if validate:
            validate_model(self)

    @property
    def settings(self):
        return self.schema.get("settings", {})

    def _load(self, file_path, content=None):
        """
        Loads a json/json5 file on the path

        :param file_path: file path on filesystem
        :return: parsed json
        """
        if file_path in self.included_schemas:
            loaded = self._fetch_included(file_path)
        else:
            extension = pathlib.Path(file_path).suffix.lower()[1:]
            if extension not in self.loaders:
                raise RuntimeError(
                    f"Can not load {file_path} - no loader has been found for extension {extension} "
                    f"in entry point group oarepo_model_builder.loaders"
                )
            loaded = self.loaders[extension](file_path, self, content=content)
        return loaded

    def _fetch_included(self, file_path):
        included = self.included_schemas[file_path]
        if callable(included):
            return included(self)
        extension = pathlib.Path(included).suffix.lower()[1:]
        if extension not in self.loaders:
            raise RuntimeError(
                f"Can not load {included} - no loader has been found for extension {extension} "
                f"in entry point group oarepo_model_builder.loaders"
            )
        return self.loaders[extension](included, self)

    def _load_included_file(self, location, source_locations=None):
        """
        Resolve and load an included file. Internal method called when loading schema.
        If the included file contains a json pointer,
        return only the part identified by the json pointer.

        :param file_id: the id of the included file, might contain #xpointer
        :return:    loaded json
        """
        if "#" in location:
            file_id, json_pointer_or_id = location.rsplit("#", 1)
        else:
            file_id = location
            json_pointer_or_id = None

        if not file_id or file_id == ".":
            ret = self.schema
        else:
            file_path = self._resolve_file_path(file_id, source_locations)
            if file_path:
                # relative include
                ret = self._load(file_path)
            else:
                if file_id not in self.included_schemas:
                    raise IncludedFileNotFoundException(
                        f"Included file {file_id} not found in includes"
                    )

                ret = self._fetch_included(file_id)

        if json_pointer_or_id:
            if json_pointer_or_id.startswith("/"):
                ret = resolve_pointer(ret, json_pointer_or_id)
            else:
                ret = resolve_id(ret, json_pointer_or_id)
                if not ret:
                    raise IncludedFileNotFoundException(
                        f"Element with id {json_pointer_or_id} not found in {file_id}"
                    )

        ret = copy.deepcopy(ret)
        if ret is None:
            ret = {}
        ret.pop("$id", None)
        return ret

    def _resolve_file_path(self, file_id, source_locations):
        if file_id in self.included_schemas:
            return file_id
        source_locations = [*(source_locations or []), *self.source_locations]
        for location in source_locations:
            pth = Path(location) / file_id
            if pth.exists():
                return pth
        pth = self.abs_path.parent / file_id
        if pth.exists():
            return pth
        return None

    def _resolve_shortcuts(self, element):
        if isinstance(element, dict):
            for v in element.values():
                self._resolve_shortcuts(v)
        elif isinstance(element, list):
            for v in element:
                self._resolve_shortcuts(v)

    def _resolve_references(self, element, stack, only_keys):
        resolved_count = 0
        if isinstance(element, dict):
            # find a reference and if there is one, resolve it
            modified = True
            while modified:
                modified = False
                for key in list(element.keys()):
                    if key not in only_keys:
                        continue
                    resolved_count += self._resolve_reference_key(
                        element, key, stack, only_keys
                    )
                    # it is possible that the reference introduced another reference,
                    # so try it once again
                    modified = True

            for k, v in element.items():
                resolved_count += self._resolve_references(v, stack + [k], only_keys)

        elif isinstance(element, list):
            for v in element:
                resolved_count += self._resolve_references(v, stack, only_keys)
        return resolved_count

    def _resolve_reference_key(self, element, key, stack, only_keys):
        included_name = element[key]

        # if it is a dictionary, then probably it is a name of a property,
        # so keep it
        if isinstance(included_name, dict):
            return self._resolve_references(element, stack, only_keys)

        # not a dict, so pop the key
        element.pop(key)

        if not isinstance(included_name, list):
            included_name = [included_name]
        for name in included_name:
            if not name:
                raise IncludedFileNotFoundException(
                    f"No file for use at path {'/'.join(stack)}"
                )
            self._load_and_merge_reference(element, key, name, stack)
        # it was resolved => return number of resolved references (1)
        return 1

    def _load_and_merge_reference(self, element, key, name, stack):
        included_data = self._load_included_file(name)

        context = {}
        for rp in self._reference_processors[key]:
            included_data = rp(
                included_data,
                element=element,
                key=key,
                name=name,
                context=context,
            )
        deepmerge(element, included_data, [], listmerge="keep")
        for rp in self._post_reference_processors[key]:
            rp(
                element=element,
                key=key,
                name=name,
                context=context,
            )

    @property
    def abs_path(self):
        return Path(self.file_path).absolute()

    def get_schema_section(self, profile, section, prepare_context=None):
        if not isinstance(section, (tuple, list)):
            section = (section,)
        section = tuple(section)
        key = (profile, section)
        if key in self._sections:
            return self._sections[key]
        from oarepo_model_builder.datatypes import datatypes

        data = self.schema
        for p in section:
            if p in data:
                data = data[p]
            else:
                data = {}
                break
        if "type" not in data:
            data["type"] = "model"
        parsed_section = datatypes.get_datatype(
            parent=None,
            data=data,
            key=None,
            model=data,
            schema=self,
        )
        prepare_context = prepare_context or {}
        prepare_context.setdefault("profile", profile)
        prepare_context.setdefault("profile_module", profile + "s")
        prepare_context.setdefault("profile_upper", profile.upper())
        parsed_section.prepare(prepare_context)
        self._sections[key] = parsed_section
        return parsed_section

    def _strip_ignored_elements(self):
        extension_elements = self.schema.get("settings", {}).get(
            "extension-elements", []
        )
        extension_elements = [
            *extension_elements,
            *[f"^{x}" for x in extension_elements],
        ]
        if not extension_elements:
            return

        def remove_extension_elements(d):
            if isinstance(d, dict):
                for k, v in list(d.items()):
                    if k in extension_elements:
                        del d[k]
                    elif isinstance(v, (dict, list)):
                        remove_extension_elements(v)
            elif isinstance(d, list):
                for v in d:
                    if isinstance(v, (dict, list)):
                        remove_extension_elements(v)

        remove_extension_elements(self.schema)


def resolve_id(json, element_id):
    if isinstance(json, dict):
        if "$id" in json and json["$id"] == element_id:
            return json
        continue_with = json.values()
    elif isinstance(json, (tuple, list)):
        continue_with = json
    else:
        return None
    for k in continue_with:
        ret = resolve_id(k, element_id)
        if ret is not None:
            return ret


def remove_star_keys(schema):
    if isinstance(schema, dict):
        for k, v in list(schema.items()):
            if k.startswith("*"):
                del schema[k]
            else:
                remove_star_keys(v)
    elif isinstance(schema, (list, tuple)):
        for k in schema:
            remove_star_keys(k)


def use_star_keys(schema):
    if isinstance(schema, dict):
        for k, v in list(schema.items()):
            if k.startswith("*"):
                del schema[k]
                schema[k[1:]] = v
        for v in schema.values():
            use_star_keys(v)
    elif isinstance(schema, (list, tuple)):
        for k in schema:
            use_star_keys(k)



import itertools

import importlib_metadata
import marshmallow as ma

from oarepo_model_builder.utils.import_class import import_class


def ExtensibleSchema(  # NOSONAR - title case because used inside a class
    extensibility_entry_point, *base_schemas
):
    def creator(*args, **kwargs):
        extra_validators = []
        for ep in sorted(
            importlib_metadata.entry_points(
                group=f"oarepo_model_builder.validation.{extensibility_entry_point}"
            ),
            key=lambda x: x.name,
        ):
            extra_validators.append(ep.load())
        base_schema_classes = []
        for bs in base_schemas:
            if isinstance(bs, str):
                bs = import_class(bs)
            base_schema_classes.append(bs)
        name = "_".join(
            x.__name__ for x in itertools.chain(base_schema_classes, extra_validators)
        )
        return type(name, tuple([*extra_validators, *base_schema_classes]), {})(
            *args, **kwargs
        )

    creator.__name__ = (
        base_schemas[0].__name__
        if base_schemas
        else f"ExtensibleSchema_{extensibility_entry_point}"
    )
    return creator


class ExtensibleSchemaMetaClass(type(ma.Schema)):
    def __new__(mcs, name, bases, attrs):
        return super().__new__(mcs, name, (*extra_validators, bases), attrs)



from collections import defaultdict
from typing import List

from ..datatypes.components.facets import FacetDefinition
from .invenio_base import InvenioBaseClassPythonBuilder
from .invenio_record_facets import get_distinct_facets


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search_options"
    section = "search-options"
    template = "record-search-options"

    def finish(self, **extra_kwargs):
        facets: List[FacetDefinition] = get_distinct_facets(self.current_model)
        facet_groups, default_group, sort_options = facet_data(
            facets, self.current_model
        )

        extra_kwargs["facet_groups"] = facet_groups
        extra_kwargs["default_group"] = default_group
        extra_kwargs["sort_definition"] = sort_options
        super().finish(**extra_kwargs)


def facet_data(facets, current_model):
    facet_groups = {}
    default_group = []
    search_data = []

    # gather all facet group names
    facet_group_names = set()
    for f in facets:
        facet_group_names.update(f.facet_groups.keys())

    top_level_facet_groups = current_model.definition.get("facets", {}).get(
        "facet-groups", {}
    )
    facet_group_names.update(top_level_facet_groups.keys())
    path_to_facet_group = defaultdict(dict)

    for group, group_def in top_level_facet_groups.items():
        for path, priority in group_def.items():
            path_to_facet_group[group][path] = priority

    # for each group name, gather all facets with that group name
    # sort them by their order in the group
    # and add them to the facet_groups dict
    for group in sorted(facet_group_names):
        # skip the default group
        if group == "_default":
            continue

        # gather all facets with this facet_group_name
        group_members = []
        for f in facets:
            if group in f.facet_groups:
                group_members.append((f, f.facet_groups[group]))

            # if there is a definition of facet groups on the model's facets -> add the selected facets as well
            for path, priority in path_to_facet_group.get(group, {}).items():
                if f.dot_path.startswith(path):
                    group_members.append((f, priority))

        # sort the group members by their order in the facet_group
        group_members.sort(key=lambda x: x[1])

        if group not in facet_groups.keys():
            facet_groups[group] = {}

        for f, _ in group_members:
            facet_groups[group][f.path] = "facets." + f.path

    for f in facets:
        if len(f.facet_groups) > 0:
            default_group.append({f.path: "facets." + f.path})
        search_data.append({f.path: "facets." + f.path})

    if "sortable" in current_model.definition:
        sort_options = current_model.definition["sortable"]
    else:
        sort_options = {}
    return facet_groups, default_group, sort_options



from pkg_resources import parse_version

from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.outputs.cfg import CFGOutput


class SetupCfgBuilder(OutputBuilder):
    TYPE = "setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")
        output.setdefault(
            "metadata",
            "name",
            self.current_model.definition["module"]["base"].replace("_", "-"),
        )
        version = self.schema.get("version", "1.0.0dev1")
        output.setdefault("metadata", "version", version)
        try:
            if parse_version(output.get("metadata", "version").value) < parse_version(
                version
            ):
                output.set("metadata", "version", version)
        except:
            # may be, for example, attr:..., so ignore exception
            pass
        output.setdefault(
            "metadata",
            "description",
            f"Repository model for {self.current_model.definition['model-name']}",
        )
        output.setdefault("metadata", "authors", "")

        output.setdefault("options", "python", ">=3.9")

        output.add_dependency("oarepo-runtime", ">=1.0.0")
        output.add_dependency("oarepo-global-search", ">=1.0.23")

        output.setdefault("options", "packages", "find:")

        output.setdefault(
            "options.package_data", "*", "*.json, *.rst, *.md, *.json5, *.jinja2"
        )

        if "runtime-dependencies" in self.schema:
            for dep, value in self.schema["runtime-dependencies"].items():
                if value[0] not in (">", "<", "="):
                    value = ">=" + value
                output.add_dependency(dep, value)

        if "dev-dependencies" in self.schema:
            for dep, value in self.schema["dev-dependencies"].items():
                if value[0] not in (">", "<", "="):
                    value = ">=" + value
                output.add_dependency(
                    dep, value, group="options.extras_require", section="devs"
                )



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent


class EnumComponent(DataTypeComponent):
    class ModelSchema(ma.Schema):
        enum = ma.fields.List(
            ma.fields.Raw, metadata={"doc": "A list of possible values"}
        )

    def process_marshmallow(self, datatype, section, **kwargs):
        enum = datatype.definition.get("enum")
        if enum:
            section.config.setdefault("validators", []).append(
                f"{{{{marshmallow.validate.OneOf}}}}({repr(enum)})"
            )

    def process_ui(self, datatype, section, **kwargs):
        enum = datatype.definition.get("enum")
        if enum:
            section.config.setdefault("marshmallow", {}).setdefault(
                "validators", []
            ).append(f"{{{{marshmallow.validate.OneOf}}}}({repr(enum)})")



from pathlib import Path
from typing import List, Union

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema


class Profile:
    def build(
        self,
        model: ModelSchema,
        profile: str,
        model_path: List[str],
        output_directory: Union[str, Path],
        builder: ModelBuilder,
        **kwargs,
    ):
        model.current_profile = self
        builder.build(model, profile, model_path, output_directory, **kwargs)



from collections import defaultdict
from typing import List

from oarepo_model_builder.datatypes import DataType, datatypes
from oarepo_model_builder.datatypes.components.marshmallow.graph import (
    collect_imports,
    set_package_dependencies,
    sort_by_reference_count,
)
from oarepo_model_builder.datatypes.components.marshmallow.object import (
    MarshmallowClass,
)
from oarepo_model_builder.utils.jinja import package_name

from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordMarshmallowBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_schema"
    template = "marshmallow"
    extra_imports = []
    build_class_method = "marshmallow_build_class"
    skip_if_not_generating = False

    def build_node(self, node: DataType):
        # everything is done in finish
        pass

    def finish(self, **extra_kwargs):
        classes: List[MarshmallowClass] = self._generate_classes(self.current_model)
        classes.sort(
            key=lambda x: (
                "Metadata" in x.class_name,
                "Record" in x.class_name,
                x.class_name,
            )
        )
        classes_by_packages = defaultdict(list)
        for cls in classes:
            classes_by_packages[package_name(cls.class_name)].append(cls)
        for single_package_classes in classes_by_packages.values():
            sort_by_reference_count(single_package_classes)
        if len(classes_by_packages.keys()) > 1:
            set_package_dependencies(classes_by_packages)

        for pn, single_package_classes in classes_by_packages.items():
            collect_imports(pn, single_package_classes)
            # generate and merge python source
            self.generate_package(pn, single_package_classes)
        # files were generated, so we are not calling super().finish() here

    def _generate_classes(self, node: DataType):
        classes = []
        to_process = [node]
        while to_process:
            n = to_process.pop(0)

            marshmallow_section = n.section_marshmallow
            if marshmallow_section.config.get(
                "class"
            ) and marshmallow_section.config.get("generate", True):
                datatypes.call_components(
                    n,
                    self.build_class_method,
                    classes=classes,
                )
            if marshmallow_section.config.get("generate", True):
                to_process.extend(marshmallow_section.children.values())
                if marshmallow_section.item:
                    to_process.append(marshmallow_section.item)
        return classes

    def generate_package(self, package_name, package_classes: List[MarshmallowClass]):
        python_path = self.class_to_path(f"{package_name}.Dummy")

        imports = [*self.extra_imports]

        for cls in package_classes:
            imports.extend(cls.imports)
            for fld in cls.fields:
                imports.extend(fld.imports)

        imports = list(
            sorted(set(imports), key=lambda x: (x.import_path or "", x.alias or ""))
        )

        self.process_template(
            python_path,
            self.template,
            current_module=package_name,
            imports=imports,
            generated_classes=package_classes,
        )



def convert_to_absolute_class_name(class_name, package_name):
    if "." not in class_name:
        return f"{package_name}.{class_name}"
    if class_name.startswith("."):
        package_path = package_name.split(".")
        while class_name.startswith("."):
            if package_path:
                package_path = package_path[:-1]
            class_name = class_name[1:]
        if package_path:
            class_name = f"{'.'.join(package_path)}.{class_name}"
    return class_name



from oarepo_model_builder.outputs.python import PythonOutput

from ..datatypes import Section
from ..datatypes.model import Link
from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_service_config"
    section = "service-config"
    template = "service-config"

    def process_template(self, python_path, template, **extra_kwargs):
        if self.parent_modules:
            self.create_parent_modules(python_path)
        output: PythonOutput = self.builder.get_output("python", python_path)

        links_section: Section = self.current_model.section_links
        imports = set()
        link: Link
        for s in links_section.config.values():
            for link in s:
                imports.update(link.imports)
        imports = list(imports)

        for ll in links_section.config.values():
            ll.sort(key=lambda x: x.name or "")

        context = dict(
            settings=self.settings,
            current_model=self.current_model,
            links=links_section.config,
            link_imports=imports,
            **extra_kwargs,
        )
        output.merge(template, context)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordListBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_list"
    section = "record-list"
    template = "record-list"



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.datatypes.components.model.utils import set_default
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent


class RecordSearchOptionsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be pasted to search options module"},
    )
    module = ma.fields.String(
        metadata={"doc": "Module where the facets will be placed"}
    )
    generate = ma.fields.Boolean()
    skip = ma.fields.Boolean()
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "Model base classes"},
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )

    sort_options_field = ma.fields.Str(
        attribute="sort-options-field",
        data_key="sort-options-field",
    )

    fields = ma.fields.Dict(
        keys=ma.fields.Str(),
        values=ma.fields.Str(),
        metadata={"doc": "Fields to be used in search options"},
    )


class SearchOptionsModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        search_options = ma.fields.Nested(
            RecordSearchOptionsSchema,
            required=False,
            data_key="search-options",
            attribute="search-options",
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        module = datatype.definition["module"]["qualified"]
        profile_module = context["profile_module"]
        record_search_prefix = datatype.definition["module"]["prefix"]

        record_search_options = set_default(datatype, "search-options", {})
        record_search_options.setdefault("generate", True)
        module = record_search_options.setdefault(
            "module", f"{module}.services.{profile_module}.search"
        )

        record_search_options.setdefault("extra-code", "")
        record_search_options.setdefault(
            "class", f"{module}.{record_search_prefix}SearchOptions"
        )
        record_search_options.setdefault(
            "base-classes",
            ["invenio_records_resources.services.SearchOptions{InvenioSearchOptions}"],
        )
        record_search_options.setdefault(
            "imports",
            [],
        )
        record_search_options.setdefault(
            "fields",
            {},
        )
        record_search_options.setdefault("sort-options-field", "sort_options")



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_resource_config"
    section = "resource-config"
    template = "resource-config"

    def finish(self, **extra_kwargs):
        profile = self.current_model.profile
        super().finish(profile=profile, **extra_kwargs)


import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import parent_module
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .service import ServiceModelComponent
from .utils import set_default


class ModelPermissionsSchema(ma.Schema):
    generate = ma.fields.Bool(
        metadata={"doc": "Set to true to generate the permissions class"}
    )
    presets = ma.fields.List(
        ma.fields.String(),
        metadata={
            "doc": "A list of presets that will be merged with the permissions class."
        },
    )
    class_ = ma.fields.String(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the generated class"},
    )
    base_classes = ma.fields.List(
        ma.fields.String(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "Base classes for the generated permission class"},
    )
    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be copied below the permission class"},
    )
    module = ma.fields.String(
        metadata={"doc": "Module where the permissions will be placed"}
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()

    class Meta:
        unknown = ma.RAISE


class PermissionsModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, ServiceModelComponent]

    class ModelSchema(ma.Schema):
        permissions = ma.fields.Nested(
            ModelPermissionsSchema(),
            required=False,
            metadata={"doc": "Permissions settings"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
        record_services_module = parent_module(datatype.definition["service"]["module"])

        permissions = set_default(datatype, "permissions", {})
        permissions.setdefault("generate", True)
        permissions.setdefault("presets", ["everyone"])
        permissions.setdefault("extra-code", "")
        permissions_module = permissions.setdefault(
            "module",
            f"{record_services_module}.permissions",
        )
        permissions.setdefault(
            "class",
            f"{permissions_module}.{prefix}PermissionPolicy",
        )
        permissions.setdefault(
            "base-classes",
            ["invenio_records_permissions.RecordPermissionPolicy"],
        )
        permissions.setdefault(
            "imports",
            [],
        )
